# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.state_machine import StateMachine
from mitzasql.db.connection import Connection
from mitzasql.db.model import (DBTablesModel, TriggerModel, ProcedureModel, TableInfoModel, ViewInfoModel)
from mitzasql.ui.screens.screen import Screen
from mitzasql.ui.widgets.db_view import DBView
from mitzasql.ui.widgets.db_tables_view import DBTablesView
from mitzasql.ui.widgets.table_view import TableView
from mitzasql.ui.widgets.query_view import QueryView
from mitzasql.ui.widgets.query_log_widget import QueryLogWidget
from mitzasql.ui.widgets.session_popup_launcher import SessionPopupLauncher
from mitzasql.ui.widgets.trigger_widget import TriggerWidget
from mitzasql.ui.widgets.procedure_widget import ProcedureWidget
from mitzasql.ui.widgets.row_widget import RowWidget
from mitzasql.ui.widgets.view_info_widget import ViewInfoWidget
from mitzasql.ui.widgets.table_info_widget import TableInfoWidget
from mitzasql.ui.widgets.table_changer_widget import TableChangerWidget
from mitzasql.ui.widgets.help_widget import HelpWidget
from .widgets_factory import WidgetsFactory
from . import states

class Session(Screen):
    def __init__(self, connection):
        super().__init__()
        self.view = SessionPopupLauncher(urwid.SolidFill(' '))
        self._connection = connection
        self._last_database = None
        self._last_table = None
        self._last_query = None
        self._last_primary_view = None
        self._state_machine = self._init_state_machine()
        self._state_machine.set_initial_state(states.STATE_INITIAL)
        self._widgets_factory = WidgetsFactory(self._state_machine)
        self._state_machine.run()

    def _init_state_machine(self):
        state_machine = StateMachine('session')
        state_machine.add_state(states.STATE_INITIAL, self.initial_state, {
            'database_selected': states.STATE_SHOW_DB_TABLES,
            'no_database_selected': states.STATE_SHOW_DATABASES
            })

        state_machine.add_state(states.STATE_SHOW_DATABASES, self.show_databases, {
            'select': states.STATE_SHOW_DB_TABLES,
            'quit': states.STATE_QUIT,
            'run_query': states.STATE_SHOW_QUERY_TABLE
            })

        state_machine.add_state(states.STATE_SHOW_DB_TABLES,
                self.show_db_tables, {
            'back': states.STATE_SHOW_DATABASES,
            'select': states.STATE_SHOW_TABLE,
            'quit': states.STATE_QUIT,
            'run_query': states.STATE_SHOW_QUERY_TABLE
            })

        state_machine.add_state(states.STATE_SHOW_TABLE, self.show_table, {
            'back': states.STATE_SHOW_DB_TABLES,
            'quit': states.STATE_QUIT,
            'run_query': states.STATE_SHOW_QUERY_TABLE
            })

        state_machine.add_state(states.STATE_SHOW_QUERY_TABLE, self.show_query_table, {
            'run_query': states.STATE_SHOW_QUERY_TABLE,
            'back': states.STATE_GOTO_PREV_VIEW,
            'quit': states.STATE_QUIT,
            })
        state_machine.add_state(states.STATE_GOTO_PREV_VIEW,
                self.goto_prev_view, {
                    'show_db_view': states.STATE_SHOW_DATABASES,
                    'show_db_tables_view': states.STATE_SHOW_DB_TABLES,
                    'show_table_view': states.STATE_SHOW_TABLE
                    })
        state_machine.add_state(states.STATE_QUIT, self.quit, {
            'back': states.STATE_GOTO_PREV_VIEW
            })
        return state_machine

    def quit(self, *args, **kwargs):
        def goto_previous_state(*args, **kwargs):
            self._state_machine.change_state('back')
        return self.view.show_quit_dialog(on_no=goto_previous_state)

    def initial_state(self):
        db_name = self._connection.database
        if db_name is None or len(db_name) == 0:
            self._state_machine.change_state('no_database_selected')
            return
        self._state_machine.change_state('database_selected', self, db_name)

    def show_databases(self, *args, **kwargs):
        self.focused_widget = self._widgets_factory.create('databases_view', self._connection)
        self._bind_help_handler(self.focused_widget)
        self._bind_log_handler(self.focused_widget)
        self.view.original_widget = self.focused_widget
        self._last_primary_view = self.focused_widget
        self.focused_widget.set_model_error_handler(self.handle_model_error)

    def show_db_tables(self, emitter, database=None):
        if database is None:
            database = self._last_database
        self._last_database = database
        self.focused_widget = self._widgets_factory.create('db_tables_view', database, self._connection)
        self._bind_help_handler(self.focused_widget)
        self._bind_log_handler(self.focused_widget)
        self.view.original_widget = self.focused_widget
        self._last_primary_view = self.focused_widget
        self.focused_widget.set_model_error_handler(self.handle_model_error)

        if hasattr(self.focused_widget, 'select_handler_bound'):
            return

        non_table_types = ('PROCEDURE', 'TRIGGER', 'FUNCTION')
        def on_select(emitter, row):
            object_type = row[-1]
            if object_type not in non_table_types:
                self._state_machine.change_state('select', emitter, row[0])
                return

            if object_type == 'TRIGGER':
                model = TriggerModel(self._connection, self._connection.database, row[0])
                if model.last_error is not None:
                    self.view.show_error(model.last_error)
                    return
                widget = TriggerWidget(model)
                self.view.show_big_popup(widget)
                return

            if object_type == 'FUNCTION' or object_type == 'PROCEDURE':
                model = ProcedureModel(self._connection, self._connection.database, row[0])
                if model.last_error is not None:
                    self.view.show_error(model.last_error)
                    return
                widget = ProcedureWidget(model)
                self.view.show_big_popup(widget)
                return

        urwid.connect_signal(self.focused_widget, self.focused_widget.SIGNAL_ACTION_SELECT_TABLE, on_select)
        self.focused_widget.select_handler_bound = True

    def show_table(self, emitter, table):
        self.focused_widget = self._widgets_factory.create('table_view', table, self._connection)
        self.view.original_widget = self.focused_widget
        self._bind_help_handler(self.focused_widget)
        self._bind_log_handler(self.focused_widget)
        self._last_primary_view = self.focused_widget
        self.focused_widget.set_model_error_handler(self.handle_model_error)

        if not hasattr(self.focused_widget, 'change_table_handler_bound'):
            def on_change_table(emitter):
                database = self._connection.database
                model = DBTablesModel(self._connection, database)
                if model.last_error is not None:
                    self.view.show_error(model.last_error)
                    return
                widget = TableChangerWidget(model)
                urwid.connect_signal(widget, widget.SIGNAL_CHANGE_TABLE,
                        self.switch_table)
                self.view.show_table_changer(widget)
                return

            urwid.connect_signal(self.focused_widget,
                    self.focused_widget.SIGNAL_ACTION_CHANGE_TABLE,
                    on_change_table)
            self.focused_widget.change_table_handler_bound = True


        if not hasattr(self.focused_widget, 'select_handler_bound'):
            def on_select(emitter, row):
                columns = self.focused_widget.model.columns
                widget = RowWidget(row, columns)
                self.view.show_big_popup(widget)
                return

            urwid.connect_signal(self.focused_widget,
                    self.focused_widget.SIGNAL_ACTION_SELECT_ROW, on_select)
            self.focused_widget.select_handler_bound = True

        if not hasattr(self.focused_widget, 'info_handler_bound'):
            def on_info(emitter, action):
                table_name = emitter.model.table_name

                result = TableInfoModel.is_view(self._connection,
                        self._connection.database, table_name)

                if result is not False:
                    viewInfoModel = result
                    if viewInfoModel.last_error is not None:
                        self.view.show_error(viewInfoModel.last_error)
                        return
                    widget = ViewInfoWidget(viewInfoModel)
                    self.view.show_big_popup(widget)
                    return
                else:
                    model = TableInfoModel(self._connection,
                            self._connection.database, table_name)
                    if model.last_error is not None:
                        self.view.show_error(model.last_error)
                        return
                    widget = TableInfoWidget(model)
                    self.view.show_big_popup(widget)
                    return

            urwid.connect_signal(self.focused_widget,
                    self.focused_widget.SIGNAL_ACTION_INFO, on_info)
            self.focused_widget.info_handler_bound = True

    def show_query_table(self, emitter, query):
        if not isinstance(emitter, QueryView):
            self.view.show_loading_dialog()

        self.focused_widget = self._widgets_factory.create('query_view', query,
                self._connection.fresh)

        self._bind_help_handler(self.focused_widget)
        self._bind_log_handler(self.focused_widget)
        self.view.original_widget = self.focused_widget
        if not isinstance(emitter, QueryView):
            self.view.close_pop_up()
        self.focused_widget.set_model_error_handler(self.handle_model_error)

        # If query cursor did not return rows, show info message
        model = self.focused_widget.model
        if not model.last_error and not model.has_rows:
            self.view.show_info(u'Affected rows: {0}'.format(model.affected_rows))
            return

        if hasattr(self.focused_widget, 'select_handler_bound'):
            return

        def on_select(emitter, row):
            columns = self.focused_widget.model.columns
            widget = RowWidget(row, columns)
            self.view.show_big_popup(widget)
            return

        urwid.connect_signal(self.focused_widget,
                self.focused_widget.SIGNAL_ACTION_SELECT_ROW, on_select)
        self.focused_widget.select_handler_bound = True

    def goto_prev_view(self, *args, **kwargs):
        if isinstance(self._last_primary_view, DBView):
            self._state_machine.change_state('show_db_view', self)
            return

        if isinstance(self._last_primary_view, DBTablesView):
            self._state_machine.change_state('show_db_tables_view', self)
            return

        if isinstance(self._last_primary_view, TableView):
            self._state_machine.change_state('show_table_view', self,
                    self._last_primary_view.table)
            return

    def handle_model_error(self, emitter, model, error):
        if self._connection.is_fatal_error(error):
            return self.view.show_fatal_error(error)

        return self.view.show_error(error)

    def switch_table(self, emitter, table):
        self.view.close_pop_up()
        self.show_table(emitter, table)

    def _bind_help_handler(self, view):
        if hasattr(view, 'help_handler_bound'):
            return

        def on_help(emitter, action):
            widget = HelpWidget()
            self.view.show_big_popup(widget)

        view.help_handler_bound = True
        urwid.connect_signal(view, view.SIGNAL_ACTION_HELP, on_help)

    def _bind_log_handler(self, view):
        if hasattr(view, 'log_handler_bound'):
            return

        def on_show_log(emitter, action):
            widget = QueryLogWidget(self._connection)
            self.view.show_big_popup(widget)

        view.log_handler_bound = True
        urwid.connect_signal(view, view.SIGNAL_ACTION_QUERY_LOG, on_show_log)




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import urwid

from mitzasql.ui.widgets.info_widget import InfoWidget

class HelpWidget(InfoWidget):

    def __init__(self):
        path = os.path.join(os.path.dirname(__file__), 'help.txt')
        try:
            with open(path) as file:
                help = file.read()
        except:
            help = u'Error: Unable to open help file'
        contents = [urwid.AttrMap(urwid.Text(help), 'default'),
                # Empty line is required, or else the 'end' key won't work
                urwid.AttrMap(urwid.Text(''), 'default')]
        super().__init__(contents)

    @property
    def name(self):
        return u'Help'



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import urwid

from mitzasql.ui.widgets.info_widget import InfoWidget

class QueryLogWidget(InfoWidget):
    def __init__(self, connection):
        contents = self._create_log(connection.QUERY_LOG)
        super().__init__(contents)
        self.focus_position = len(connection.QUERY_LOG) * 2

    @property
    def name(self):
        return u'Query Log'

    def _create_log(self, log):
        contents = []
        for entry in log:
            date, query, params, duration = entry
            date = date.strftime('%a %H:%M:%S')
            duration = '{:.3f}'.format(duration)
            log_entry = '{0}: {1}\nParams: {2}\nDuration: {3}s'.format(date,
                    query, params, duration)

            contents.append(urwid.AttrMap(urwid.Text(log_entry), 'default'))
            contents.append(urwid.Divider('-'))

        contents.append(urwid.Text(''))
        return contents



import tempfile
import os
import pytest

from mitzasql.ui.screens.session_select import session_select
from mitzasql.ui.widgets.saved_sessions_list import SavedSessionsList
from mitzasql.ui.widgets.create_new_session_dialog import CreateNewSessionDialog
from mitzasql.ui.widgets.edit_session_form import EditSessionForm
from mitzasql.sessions_registry import SessionsRegistry

screen_size = (120, 30)
bad_connection_data_set = [
        {
            'data': {},
            'expected_message': u'Session name is required'
        },
        {
            'data': {
                'name': ''
                },
            'expected_message': u'Session name is required'
        },
        {
            'data': {
                'name': 'some name',
                },
            'expected_message': u'Host is required'
        },
        {
            'data': {
                'name': 'some name',
                'host': ''
                },
            'expected_message': u'Host is required'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'unknown protocol'
                },
            'expected_message': u'Host is missing protocol (tcp://, unix://)'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'tcp://'
                },
            'expected_message': u'Host must not be empty'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'unix://'
                },
            'expected_message': u'Host must not be empty'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'tcp://localhost'
                },
            'expected_message': u'Port is required'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'tcp://localhost',
                'port': ''
                },
            'expected_message': u'Port is required'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'tcp://localhost',
                'port': '3306'
                },
            'expected_message': u'Username is required'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'tcp://localhost',
                'port': '3306',
                'username': ''
                },
            'expected_message': u'Username is required'
        },
        {
            'data': {
                'name': 'some name',
                'host': 'unix:///var/run/mysql.sock',
                'username': ''
                },
            'expected_message': u'Username is required'
        },
        ]

@pytest.fixture
def registry_with_a_session():
    sessions_file = tempfile.NamedTemporaryFile()
    registry = SessionsRegistry(sessions_file.name)
    registry.add({
        'name': 'Test session',
        'host': 'tcp://localhost',
        'port': '3306',
        'username': 'root',
        'password': 'root',
        'database': ''
        })
    return registry

@pytest.fixture
def empty_registry():
    sessions_file = tempfile.NamedTemporaryFile()
    registry = SessionsRegistry(sessions_file.name)
    return registry

@pytest.fixture(params=bad_connection_data_set)
def bad_connection_data(request):
    return request.param

@pytest.fixture
def good_connection_data():
    return {
            'name': 'Tcp connection',
            'host': 'tcp://localhost',
            'port': '3306',
            'username': 'root',
            'password': 'root',
            'database': 'mysql'
            }

def test_screen_displays_sessions_list(registry_with_a_session):
    sessions_select_screen = session_select.SessionSelect(registry_with_a_session)
    assert isinstance(sessions_select_screen.view.original_widget.body,
            SavedSessionsList) is True

def test_screen_displays_create_new_session_dialog(empty_registry):
    sessions_select_screen = session_select.SessionSelect(empty_registry)
    assert isinstance(sessions_select_screen.view.original_widget.body.original_widget,
            CreateNewSessionDialog) is True

def test_screen_shows_the_create_new_session_form(empty_registry):
    sessions_select_screen = session_select.SessionSelect(empty_registry)
    sessions_select_screen.view.keypress(screen_size, 'y')
    assert isinstance(sessions_select_screen.view.original_widget.body,
            EditSessionForm) is True


def test_screen_shows_form_validation_error_on_create_new_session(empty_registry, bad_connection_data):
    sessions_select_screen = session_select.SessionSelect(empty_registry)

    sessions_select_screen.view.keypress(screen_size, 'y')

    form = sessions_select_screen._edit_form_widget
    form.refresh(bad_connection_data['data'])
    form.keypress((screen_size[0],), 'f1')

    assert form._status_bar.text == bad_connection_data['expected_message']

def test_screen_adds_new_session(empty_registry, good_connection_data):
    sessions_select_screen = session_select.SessionSelect(empty_registry)

    sessions_select_screen.view.keypress(screen_size, 'y')

    form = sessions_select_screen._edit_form_widget
    form.refresh(good_connection_data)
    form.keypress((screen_size[0],), 'f1')

    assert empty_registry.is_empty() is False
    assert isinstance(sessions_select_screen.view.original_widget.body,
            SavedSessionsList) is True

def test_screen_edits_session(registry_with_a_session):
    sessions_select_screen = session_select.SessionSelect(registry_with_a_session)

    sessions_list_widget = sessions_select_screen.view.original_widget.body
    sessions_list_widget.keypress((screen_size[0], ), 'e')

    form_data = registry_with_a_session['Test session']
    form_data['name'] = 'Test session'
    form = sessions_select_screen._edit_form_widget
    form.refresh(form_data, False)

    form._input_elements['name'].edit_text = 'Edited session'

    form.keypress((screen_size[0],), 'f1')
    assert registry_with_a_session.has_session('Edited session')
    assert len(registry_with_a_session.sessions) == 1


def test_screen_deletes_session(registry_with_a_session):
    sessions_select_screen = session_select.SessionSelect(registry_with_a_session)

    sessions_list_widget = sessions_select_screen.view.original_widget.body
    sessions_list_widget.keypress((screen_size[0], ), 'd')

    assert isinstance(sessions_select_screen.view.original_widget.body.original_widget,
            CreateNewSessionDialog) is True
    assert registry_with_a_session.is_empty() is True

def test_screen_shows_form_validation_error_on_edit(registry_with_a_session, bad_connection_data):
    sessions_select_screen = session_select.SessionSelect(registry_with_a_session)

    sessions_list_widget = sessions_select_screen.view.original_widget.body

    sessions_list_widget.keypress((screen_size[0], ), 'e')

    form = sessions_select_screen._edit_form_widget
    form.refresh(bad_connection_data['data'], False)

    form.keypress((screen_size[0],), 'f1')
    assert isinstance(sessions_select_screen.view.original_widget.body,
            EditSessionForm) is True
    assert form._status_bar.text == bad_connection_data['expected_message']



import pytest

from mitzasql.ui.widgets.cmd_proc import Autocomplete

def test_autocomplete_returns_suggestions():
    suggestions = [
            'john',
            'johnny',
            'jake',
            'jane'
            ]

    autocomplete = Autocomplete(suggestions, 'joh')

    suggestion = autocomplete.suggestion('forward')
    assert suggestion == 'john'
    suggestion = autocomplete.suggestion('forward')
    assert suggestion == 'johnny'
    suggestion = autocomplete.suggestion('forward')
    assert suggestion == 'john'
    suggestion = autocomplete.suggestion('forward')
    assert suggestion == 'johnny'

    suggestion = autocomplete.suggestion('back')
    assert suggestion == 'john'
    suggestion = autocomplete.suggestion('back')
    assert suggestion == 'johnny'







# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from collections import OrderedDict

import urwid
import mysql.connector.errors as errors

import mitzasql.db.schema as schema

MAX_ROWS_FOR_COLUMN_LENGTH_DETECTION = 100

class Model:
    SIGNAL_PRE_LOAD = 'preload'
    SIGNAL_LOAD = 'load'
    SIGNAL_NEW_DATA = 'new_data'

    SIGNALS = [SIGNAL_PRE_LOAD, SIGNAL_NEW_DATA, SIGNAL_LOAD]

    ''' Base model '''
    def __init__(self, data = [], columns = {}):
        self.data = data
        self.columns = columns
        self.rowcount = len(data)
        urwid.register_signal(self.__class__, self.SIGNALS)

    def __getitem__(self, key):
        return self.data[key]

    def __len__(self):
        return self.rowcount

    def __iter__(self):
        for row in self.data:
            yield row

    def col_index(self, name):
        for index, item in enumerate(self.columns):
            if item['name'].lower() == name.lower():
                return index
        raise IndexError()

    def search(self, keyword, col_index, pos=0, reverse=False):
        if reverse is True:
            row_index = 1
            start = 0
            end = pos
        else:
            row_index = 0
            start = pos
            end = None

        data = self.data[start:end]

        if reverse is True:
            data = reversed(data)

        for row in data:
            col = row[col_index]
            if str(keyword).lower() in str(col).lower():
                if not reverse:
                    return (start + row_index, row)
                return (end - row_index, row)

            row_index += 1

class MysqlModel(Model):
    SIGNAL_ERROR = 'error'

    '''Base mysql query model'''
    def __init__(self, connection):
        self._con = connection
        self.last_error = None
        self.SIGNALS.extend([self.SIGNAL_ERROR])
        super().__init__()
        self._fetch_data()

    def _fetch_data(self):
        cursor = self._query_db()
        if cursor is None:
            self.data = []
            self.columns = []
            self.rowcount = 0
            return False
        self.data = cursor.fetchall()
        self.columns = self._schema(cursor).columns
        self.rowcount = cursor.rowcount
        return True

    def _schema(self, cursor):
        return schema.QuerySchema(cursor,
                self.data[0:MAX_ROWS_FOR_COLUMN_LENGTH_DETECTION])

    def reload(self):
        urwid.emit_signal(self, self.SIGNAL_PRE_LOAD, self)
        self._fetch_data()
        urwid.emit_signal(self, self.SIGNAL_LOAD, self)

    def execute_query(self, query, params=None):
        try:
            cursor = self._con.query(query, params)
            self.last_error = None
            return cursor
        except errors.Error as e:
            self.last_error = e
            urwid.emit_signal(self, self.SIGNAL_ERROR, self, e)

    def _query_db(self):
        '''Creates and executes a query cursor. Must be implemented by child
        models
        '''
        raise RuntimeError("Child model must implement this method")

class DatabasesModel(MysqlModel):
    '''Model for the 'SHOW DATABASES' query'''
    def __init__(self, connection):
        super().__init__(connection)

    def _query_db(self):
        query = 'SHOW DATABASES'
        cursor = self.execute_query(query)
        return cursor

class DBTablesModel(MysqlModel):
    '''Model for all the tables in the current database'''
    def __init__(self, connection, database):
        self.database = database
        super().__init__(connection)

    def _query_db(self):
        try:
            self._con.change_db(self.database)
        except errors.Error as e:
            self.last_error = e
            urwid.emit_signal(self, self.SIGNAL_ERROR, self, e)
            return

        query = []

        # Get tables
        query.append('''
        SELECT
            TABLE_NAME AS Name,
            TABLE_ROWS AS `Rows`,
            DATA_LENGTH AS Size,
            CREATE_TIME AS Created,
            UPDATE_TIME AS Updated,
            ENGINE AS Engine,
            TABLE_COMMENT AS Comment,
            TABLE_TYPE AS Type
        FROM
            INFORMATION_SCHEMA.TABLES
        WHERE TABLE_SCHEMA = '{0}'
        '''.format(self.database))

        # Get triggers
        query.append('''
            SELECT
                TRIGGER_NAME AS Name,
                NULL AS `Rows`,
                NULL AS Size,
                CREATED AS Created,
                NULL AS Updated,
                NULL AS Engine,
                CONCAT_WS(' ', ACTION_TIMING, EVENT_MANIPULATION, 'in',
                EVENT_OBJECT_TABLE) AS Comment,
                'TRIGGER' AS Type
            FROM INFORMATION_SCHEMA.TRIGGERS
            WHERE TRIGGER_SCHEMA = '{0}'
        '''.format(self.database))

        # Get functions & procedures
        query.append('''
            SELECT
                SPECIFIC_NAME AS Name,
                NULL AS `Rows`,
                NULL AS Size,
                CREATED AS Created,
                LAST_ALTERED AS Updated,
                NULL AS Engine,
                CONVERT(ROUTINE_COMMENT USING utf8) AS Comment,
                ROUTINE_TYPE AS Type
            FROM
                INFORMATION_SCHEMA.ROUTINES
            WHERE ROUTINE_SCHEMA = '{0}'
        '''.format(self.database))
        query = '(' + '\n) UNION (\n'.join(query) + ')'
        query = 'SELECT * FROM ({0}) AS info ORDER BY Name'.format(query)
        cursor = self.execute_query(query)
        return cursor

class TableModel(MysqlModel):
    PAGINATION_LIMIT = 100

    '''Model for a database table/view'''
    def __init__(self, connection, table_name):
        self.table_name = table_name
        self.schema_error = None

        try:
            self._table_schema = schema.TableSchema(connection, table_name)
        except errors.Error as e:
            self.schema_error = e
            self.last_error = e

        self._limit = self.PAGINATION_LIMIT
        self._page = 1
        self._offset = 0
        self._column_order = None
        self._order_dir = None
        self._where = None
        super().__init__(connection)

        self.loaded_rowcount = len(self.data)
        self.rowcount = self._count_rows()

    def _fetch_data(self):
        result = super()._fetch_data()
        return result

    def reload(self, reset_limit=True, reset_order=True, reset_where=True):
        urwid.emit_signal(self, self.SIGNAL_PRE_LOAD, self)

        try:
            self._table_schema = schema.TableSchema(self._con, self.table_name)
            self.schema_error = None
            self.last_error = None
        except errors.Error as e:
            self.schema_error = e
            self.last_error = e
            urwid.emit_signal(self, self.SIGNAL_LOAD, self)
            return

        if reset_limit:
            self._page = 1
            self._offset = 0

        if reset_order:
            self._column_order = None
            self._order_dir = None

        if reset_where:
            self._where = None

        if self._fetch_data() is False:
            self.loaded_rowcount = 0
            self.rowcount = 0
            urwid.emit_signal(self, self.SIGNAL_LOAD, self)
            return

        self.loaded_rowcount = len(self.data)
        self.rowcount = self._count_rows()
        urwid.emit_signal(self, self.SIGNAL_LOAD, self)

    def _schema(self, cursor):
        self._table_schema.data = self.data[0:MAX_ROWS_FOR_COLUMN_LENGTH_DETECTION]
        self._table_schema.cursor = cursor
        return self._table_schema

    def _count_rows(self):
        query = 'SELECT COUNT(*) AS total FROM {0}'.format(self.table_name)

        if self._where:
            query += ' WHERE {0}'.format(self._where)
        cursor = self.execute_query(query)
        if cursor is None:
            return 0
        data = cursor.fetchall()
        return data[0][0]

    def load_next_set(self):
        self._increment_page()

        urwid.emit_signal(self, self.SIGNAL_PRE_LOAD, self)
        result = self._fetch_more_rows()
        if result is None:
            self._increment_page(-1)
            data = []
            data_length = 0
        else:
            data, data_length = result
        urwid.emit_signal(self, self.SIGNAL_NEW_DATA, self, data, data_length)

    def sort(self, column, direction):
        self._column_order = column
        self._order_dir = direction
        self._page = 1
        self._offset = 0

        self.reload(reset_limit=False, reset_order=False, reset_where=False)

    def filter(self, where):
        if where is None:
            self.reload(reset_limit=True, reset_order=False, reset_where=True)
            return
        self._where = where
        self.reload(reset_limit=True, reset_order=False, reset_where=False)

    def load_more(self, count):
        offset = self.loaded_rowcount
        limit = count + 100 # Load 100 more rows in advance

        old_limit = self._limit
        old_offset = self._offset

        self._limit = limit
        self._offset = offset

        urwid.emit_signal(self, self.SIGNAL_PRE_LOAD, self)
        result = self._fetch_more_rows()
        if result is None:
            self._limit = old_limit
            self._offset = old_offset
            data = []
            data_length = 0
        else:
            data, data_length = result
        urwid.emit_signal(self, self.SIGNAL_NEW_DATA, self, data, data_length)

        self._page = self.loaded_rowcount // self.PAGINATION_LIMIT
        self._limit = self.PAGINATION_LIMIT

    def _fetch_more_rows(self):
        cursor = self._query_db()
        if cursor is None:
            return None
        data = cursor.fetchall()
        data_length = len(data)
        self.data.extend(data)
        self.loaded_rowcount += data_length
        return (data, data_length)

    def _increment_page(self, value=1):
        self._page += value
        self._offset = (self._page - value) * self._limit

    def _query_db(self):
        if self.schema_error is not None:
            return
        select_columns = []
        for column, info in self._table_schema:
            if self._table_schema.column_is_text(info['type']) and info['max_len'] is not None and info['max_len'] > 256:
                column = 'LEFT(`{0}`, 256) as {1}'.format(column, column)
            elif self._table_schema.column_is_spatial(info['type']):
                column = 'ST_AsText(`{0}`) as {1}'.format(column, column)
            else:
                column = '`{0}`'.format(column)
            select_columns.append(column)

        query = 'SELECT {0} FROM `{1}`'.format(','.join(select_columns),
                self.table_name)

        if self._where:
            query += ' WHERE {0}'.format(self._where)

        if self._column_order is not None and self._order_dir is not None:
            query += ' ORDER BY `{0}` {1}'.format(self._column_order, self._order_dir)

        query += ' LIMIT {0:d}, {1:d}'.format(self._offset, self._limit)
        cursor = self.execute_query(query)
        return cursor

class QueryModel(MysqlModel):
    def __init__(self, connection, query):
        self.query = query
        self.affected_rows = 0
        self.has_rows = False
        super().__init__(connection)

    def _query_db(self):
        cursor = self.execute_query(self.query)
        if cursor is not None:
            self.has_rows = cursor.with_rows
        return cursor

    def _fetch_data(self):
        cursor = self._query_db()
        if cursor is None:
            self.data = []
            self.columns = []
            self.rowcount = 0
            self.affected_rows = 0
            return cursor is not None

        if cursor.with_rows:
            self.data = cursor.fetchall()
            self.columns = self._schema(cursor).columns
            self.rowcount = cursor.rowcount
            self.affected_rows = 0
        else:
            self.data = []
            self.columns = []
            self.rowcount = 0
            self.affected_rows = cursor.rowcount

        return True

class TriggerModel:
    def __init__(self, connection, database, name):
        self._con = connection
        self._database = database
        self._name = name
        self.last_error = None
        self.data = self._query()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def _query(self):
        query = '''
        SELECT
            *
        FROM INFORMATION_SCHEMA.TRIGGERS
        WHERE
            TRIGGER_SCHEMA = '{0}'
            AND
            TRIGGER_NAME = '{1}'
        '''.format(self._database, self._name)

        try:
            cursor = self._con.query(query, dictionary=True)
            data = cursor.fetchall()
            if data:
                return data[0]
            else:
                return {}
        except errors.Error as e:
            self.last_error = e
            return {}

class ProcedureModel:
    def __init__(self, connection, database, name):
        self._con = connection
        self._database = database
        self._name = name
        self.last_error = None
        self.data = self._query()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def _query(self):
        query = '''
        SELECT
            *,
            (
                SELECT
                    GROUP_CONCAT(PARAMETER_MODE, ' ', PARAMETER_NAME, ' ',
                    DTD_IDENTIFIER)
                FROM INFORMATION_SCHEMA.PARAMETERS AS p
                WHERE p.SPECIFIC_NAME = ROUTINE_NAME
                AND ORDINAL_POSITION > 0
                ORDER BY ORDINAL_POSITION ASC
            ) AS param_list
        FROM INFORMATION_SCHEMA.ROUTINES
        WHERE
            ROUTINE_SCHEMA = '{0}'
            AND
            SPECIFIC_NAME = '{1}'
        '''.format(self._database, self._name)

        try:
            cursor = self._con.query(query, dictionary=True)
            data = cursor.fetchall()
            if data:
                return data[0]
            else:
                return {}
        except errors.Error as e:
            self.last_error = e
            return {}

class ViewInfoModel:
    def __init__(self, connection, database, name):
        self._con = connection
        self._database = database
        self._name = name
        self.last_error = None
        self.data = self._query()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def _query(self):
        query = '''
        SELECT
            *
        FROM INFORMATION_SCHEMA.VIEWS
        WHERE
            TABLE_SCHEMA = '{0}'
            AND
            TABLE_NAME = '{1}'
        '''.format(self._database, self._name)

        try:
            cursor = self._con.query(query, dictionary=True)
            data = cursor.fetchall()
            if data:
                return data[0]
            else:
                return {}
        except errors.Error as e:
            self.last_error = e
            return {}

class TableInfoModel:
    def __init__(self, connection, database, name):
        self._con = connection
        self._database = database
        self._name = name
        self.last_error = None
        self.data = self._query()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def _query(self):
        table_data = self._get_table_data()
        if self.last_error is not None or not len(table_data):
            return {}

        create_code = self._get_create_code()
        if self.last_error is not None:
            return {}

        data = {
                'table': table_data,
                'create_code': create_code
                }
        return data

    def _get_table_data(self):
        query = '''
        SELECT
            *
        FROM INFORMATION_SCHEMA.TABLES
        WHERE
            TABLE_SCHEMA = '{0}'
            AND
            TABLE_NAME = '{1}'
        '''.format(self._database, self._name)

        try:
            cursor = self._con.query(query, dictionary=True)
            data = cursor.fetchall()
            if data:
                return data[0]
            else:
                return {}
        except errors.Error as e:
            self.last_error = e
            return {}

    def _get_create_code(self):
        query = '''
        SHOW CREATE TABLE `{0}`
        '''.format(self._name)

        try:
            cursor = self._con.query(query, dictionary=True)
            data = cursor.fetchall()
            if data:
                return data[0]['Create Table']
            else:
                return {}
        except errors.Error as e:
            self.last_error = e
            return {}

    @classmethod
    def is_view(cls, connection, database, name):
        viewInfoModel = ViewInfoModel(connection, database, name)
        if len(viewInfoModel):
            return viewInfoModel
        return False





import os
import pytest

import urwid
import mysql.connector.errors as errors
from mitzasql.db.model import QueryModel
from .connection_fixture import sakila_connection

def test_model_fetches_data(sakila_connection):
    query = 'SELECT address_id, address, location FROM address LIMIT 10'
    model = QueryModel(sakila_connection, query)
    assert model.last_error is None
    assert len(model) == 10
    assert len(model.columns) == 3
    assert 'address_id' in [c['name'] for c in model.columns]
    assert 'address' in [c['name'] for c in model.columns]
    assert 'location' in [c['name'] for c in model.columns]

def test_model_sets_last_error_when_query_is_invalid(sakila_connection):
    query = 'SELECT unknown_column, address, location FROM address LIMIT 10'
    model = QueryModel(sakila_connection, query)
    assert isinstance(model.last_error, errors.ProgrammingError)



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import sys
import signal
import time
import threading
import urwid
import re

from mitzasql.ui import theme
from mitzasql.ui import main_loop as shared_main_loop
from mitzasql.ui.builder import Builder

top_most_widget = None
palette = theme.get_palette()
builder = None
main_loop = None
simulate_input_pipe = None

def init():
    """Initialize the UI

    Creates an empty widget placeholder top most widget and initializes the
    screens builder with it
    """
    global top_most_widget, builder

    top_most_widget = urwid.WidgetPlaceholder(urwid.SolidFill(' '))
    builder = Builder(top_most_widget)

def build_screen(name, display, **kwargs):
    screen = builder.build_screen(name, display, **kwargs)
    return screen

def connect_signal(obj, name, callback, **kwargs):
    """Wrapper for urwid.connect_signal"""
    urwid.connect_signal(obj, name, callback, **kwargs)

def refresh():
    main_loop.draw_screen()

def convert_termsignals_to_exit_loop(signum, frame):
    raise urwid.ExitMainLoop()

def simulate_input(data):
    """Callback called when there's data in the main loop pipe

    Receives keys from the `simulate_input_thread` and sends input to the main
    loop

    Parameters:
    data:bytes array
    """
    if data.decode('utf-8') == 'remove_watch':
        main_loop.remove_watch_pipe(simulate_input_pipe)
        return
    main_loop.process_input([data.decode('utf-8')])

def simulate_input_thread(fd, macro_file):
    """Reads the macro file line by line and sends input to the main event loop
    through a pipe (fd)

    Parameters:
    fd:file descriptor The pipe to the main loop
    macro_file:String Path to macro file
    """
    with open(macro_file) as file:
        while True:
            line = file.readline()
            if line is '':
                break
            line = line.rstrip()
            input_type, *segments = line.split(':')
            if input_type.startswith('#'):
                continue
            if input_type == 'key':
                key = segments[0]
                if key == 'space':
                    key = ' '
                if len(segments) == 2:
                    repetitions = int(segments[1])
                else:
                    repetitions = 1
                for i in range(0, repetitions):
                    os.write(fd, key.encode('utf8'))
                    time.sleep(0.08)
            if input_type == 'key_sequence':
                sequence = ':'.join(segments)

                # Parse for environment variables placeholders
                matches = re.findall('({{.*?}})', sequence)
                for match in matches:
                    var_name = match[2:-2]
                    var_value = os.getenv(var_name, '')
                    sequence = sequence.replace(match, var_value)

                for char in sequence:
                    os.write(fd, char.encode('utf8'))
                    time.sleep(0.05)
            elif input_type == 'sleep':
                value = float(segments[0])
                time.sleep(value)
    os.write(fd, 'remove_watch'.encode('utf8'))
    time.sleep(0.1)
    os.close(fd)

def start_loop(macro_file=None):
    """Start ui event loop

    If `macro_file` is set create a new thread which reads the macro file and
    sends input to the main loop

    Parameters:
    macro_file:String Path to macrofile
    """
    global main_loop, simulate_input_pipe
    main_loop = urwid.MainLoop(top_most_widget,
            palette=palette,
            handle_mouse=False,
            pop_ups=True)
    main_loop.screen.set_terminal_properties(colors=256)

    if macro_file is not None:
        try:
            with open(macro_file) as file:
                pass
        except FileNotFoundError:
            print('Macro file does not exist!', file=sys.stderr)
            sys.exit(1)

        simulate_input_pipe = main_loop.watch_pipe(simulate_input)
        thread = threading.Thread(target=simulate_input_thread,
                args=[simulate_input_pipe, macro_file])
        thread.start()

    signal.signal(signal.SIGINT, convert_termsignals_to_exit_loop)
    signal.signal(signal.SIGQUIT, convert_termsignals_to_exit_loop)
    signal.signal(signal.SIGTERM, convert_termsignals_to_exit_loop)

    # Set a reference to the main loop in a another module
    # in order to avoid circular reference in widgets when
    # importing the ui to refresh the screen
    shared_main_loop.mutable['loop'] = main_loop
    main_loop.run()



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

import mitzasql.ui.utils as utils

class List(urwid.ListBox):
    """Custom list which emits a `selected` signal if an item is selected"""

    SIGNAL_SELECTED = 'selected'
    def __init__(self, data):
        focus_list_walker = self._create_focus_list_walker(data)
        super().__init__(focus_list_walker)
        urwid.register_signal(self.__class__, [self.SIGNAL_SELECTED])

    def _create_focus_list_walker(self, data):
        selectables = []
        data.sort()
        for item in data:
            selectable = urwid.AttrMap(urwid.SelectableIcon(item, len(item) + 1),
                    'list:item:unfocused',
                    'list:item:focused')
            selectable.item_value = item
            selectables.append(selectable)
        return urwid.SimpleFocusListWalker(selectables)

    def reset(self, data):
        self.body.clear()
        self.body = self._create_focus_list_walker(data)

    def clear(self):
        self.body.clear()

    def append(self, item):
        selectable = urwid.AttrMap(urwid.SelectableIcon(item, len(item) + 1),
                'list:item:unfocused',
                'list:item:focused')
        selectable.item_value = item
        self.body.append(selectable)

    def keypress(self, size, key):
        key = utils.vim2emacs_translation(key)
        if key == 'enter' and self.focus is not None:
            urwid.emit_signal(self, self.SIGNAL_SELECTED, self, self.focus.item_value)
            return
        return super().keypress(size, key)



import pytest
import urwid

from mitzasql.ui.widgets.table_view import TableView
from mitzasql.db.model import TableModel
from ..db.connection_fixture import sakila_connection

filters = [
        {
        'filter': ['eq', 'address_id', '1'],
        'count': 1
        },
        {
        'filter': ['neq', 'address_id', '1'],
        'count': 602
        },
        {
        'filter': ['lt', 'address_id', '10'],
        'count': 9
        },
        {
        'filter': ['lte', 'address_id', '10'],
        'count': 10
        },
        {
        'filter': ['gt', 'address_id', '600'],
        'count': 5
        },
        {
        'filter': ['gte', 'address_id', '600'],
        'count': 6
        },
        {
        'filter': ['in', 'address_id', '1,2, 3'],
        'count': 3
        },
        {
        'filter': ['nin', 'address_id', '1,2, 3'],
        'count': 600
        },
        {
        'filter': ['null', 'address2'],
        'count': 4
        },
        {
        'filter': ['nnull', 'address2'],
        'count': 599
        },
        {
        'filter': ['empty', 'address2'],
        'count': 599
        },
        {
        'filter': ['nempty', 'address2'],
        'count': 0
        },
        {
        'filter': ['like', 'address', '%hanoi%'],
        'count': 2
        },
        {
        'filter': ['nlike', 'address', '%hanoi%'],
        'count': 601
        },
        {
        'filter': ['between', 'address_id', '1', '4'],
        'count': 4
        },
        {
        'filter': ['nbetween', 'address_id', '1', '4'],
        'count': 599
        },
        {
        'filter': ['between', 'last_update', "'2014-09-25 22:29:59'", "'2014-09-25 22:30:01'"],
        'count': 4
        },
        {
        'filter': ['nbetween', 'last_update', "'2014-09-25 22:29:59'", "'2014-09-25 22:30:01'"],
        'count': 599
        },
        ]

@pytest.fixture(params=filters)
def filter_cmd(request):
    return request.param

def test_view_filters_model(sakila_connection, filter_cmd):
    model = TableModel(sakila_connection, 'address')
    view = TableView(model, sakila_connection)

    assert len(view.model) == 603

    op, *args = filter_cmd['filter']
    expected_count = filter_cmd['count']

    view.filter(op, *args)
    assert len(view.model) == expected_count

def test_view_clears_model_filter(sakila_connection):
    model = TableModel(sakila_connection, 'address')
    view = TableView(model, sakila_connection)

    assert len(view.model) == 603

    view.filter('eq', 'address_id', '1')
    assert len(view.model) == 1

    view.filter('clearfilters')
    assert len(view.model) == 603



import os
import pytest

import mysql.connector.errors as errors
from mitzasql.db.model import TableModel
from .connection_fixture import sakila_connection

table = 'actor'

where_clauses = [
        'address_id = 1',
        'address_id = "1"',
        'address_id != 1',
        'address_id < 1',
        'address_id <= 1',
        'address_id > 1',
        'address_id >= 1',
        'address_id is null',
        'address_id is not null',
        'address_id = ""',
        'address_id != ""',
        'address_id in (1, 2)',
        'address_id not in (1, 2)',
        'address_id between 1 and 10',
        'address_id not between 1 and 10',
        'address like "%hanoi%"',
        'address not like "%hanoi%"',
        ]

@pytest.fixture(params=where_clauses)
def filter(request):
    return request.param

def test_model_fetches_data(sakila_connection):
    model = TableModel(sakila_connection, table)
    model_columns = [c['name'] for c in model.columns]
    assert len(model) == 200
    assert 'actor_id' in model_columns
    assert 'first_name' in model_columns
    assert 'last_name' in model_columns
    assert 'last_update' in model_columns

    assert model.loaded_rowcount == 100

def test_model_fetches_next_records_set(sakila_connection):
    model = TableModel(sakila_connection, table)

    assert model.loaded_rowcount == 100
    model.load_next_set()
    assert model.loaded_rowcount == 200
    assert len(model) == 200

def test_model_loads_more_rows(sakila_connection):
    model = TableModel(sakila_connection, table)

    assert model.loaded_rowcount == 100
    model.load_more(50)
    assert model.loaded_rowcount == 200
    assert len(model) == 200

def test_model_sorts_rows_by_first_name_ascending(sakila_connection):
    model = TableModel(sakila_connection, table)
    model.sort('first_name', 'asc')

    row = model[0]
    assert row[0] == 132
    assert row[1] == 'ADAM'

def test_model_sorts_rows_by_first_name_descending(sakila_connection):
    model = TableModel(sakila_connection, table)
    model.sort('first_name', 'desc')

    row = model[0]
    assert row[0] == 11
    assert row[1] == 'ZERO'

def test_model_sets_last_error_when_table_doesnt_exist(sakila_connection):
    model = TableModel(sakila_connection, 'unknown table')
    assert isinstance(model.last_error, errors.ProgrammingError)

    assert len(model) == 0
    assert len(model.data) == 0
    assert model.loaded_rowcount == 0
    assert bool(model.columns) is False

def test_model_emits_error_signal(sakila_connection):
    model = TableModel(sakila_connection, table)
    assert model.last_error is None
    model.sort('non-existent-column', 'invalid-dir')
    assert isinstance(model.last_error, errors.ProgrammingError)

    assert len(model) == 0
    assert len(model.data) == 0
    assert model.loaded_rowcount == 0
    assert bool(model.columns) is False

def test_model_filters_records(sakila_connection, filter):
    model = TableModel(sakila_connection, 'address')
    model.filter(filter)

    assert model.last_error is None



import pytest

from mitzasql.ui.screens.session import (session, states)
from mitzasql.ui.widgets.db_view import DBView
from mitzasql.ui.widgets.db_tables_view import DBTablesView
from mitzasql.ui.widgets.table_view import TableView
from mitzasql.ui.widgets.query_view import QueryView
from ...db.connection_fixture import (connection, sakila_connection)

def test_screen_displays_the_databases_view(connection):
    screen = session.Session(connection)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DATABASES
    assert isinstance(screen.focused_widget, DBView)

def test_screen_transitions_from_database_view_to_db_tables_view(connection):
    screen = session.Session(connection)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DATABASES
    screen._state_machine.change_state('select', None, 'sakila')
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DB_TABLES
    assert isinstance(screen.focused_widget, DBTablesView)

def test_screens_displays_the_db_tables_view(sakila_connection):
    screen = session.Session(sakila_connection)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DB_TABLES
    assert isinstance(screen.focused_widget, DBTablesView)

def test_screen_transitions_to_table_view(sakila_connection):
    screen = session.Session(sakila_connection)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DB_TABLES
    screen._state_machine.change_state('select', None, 'actor')
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_TABLE
    assert isinstance(screen.focused_widget, TableView)

def test_screen_transitions_to_query_view(sakila_connection):
    screen = session.Session(sakila_connection)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DB_TABLES
    screen._state_machine.change_state('run_query', None, 'SELECT COUNT(*) FROM actor')
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_QUERY_TABLE
    assert isinstance(screen.focused_widget, QueryView)

def test_screen_transitions_from_query_view_to_database_view(connection):
    connection.database = None
    screen = session.Session(connection)
    assert isinstance(screen.focused_widget, DBView)
    screen._state_machine.change_state('select', None, 'sakila')
    assert isinstance(screen.focused_widget, DBTablesView)
    screen._state_machine.change_state('select', None, 'actor')
    assert isinstance(screen.focused_widget, TableView)
    screen._state_machine.change_state('run_query', None, 'SELECT COUNT(*) FROM actor')
    assert isinstance(screen.focused_widget, QueryView)
    screen._state_machine.change_state('back', screen)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_TABLE
    assert isinstance(screen.focused_widget, TableView)
    screen._state_machine.change_state('back', screen)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DB_TABLES
    assert isinstance(screen.focused_widget, DBTablesView)
    screen._state_machine.change_state('back', screen)
    assert screen._state_machine.get_current_state() == states.STATE_SHOW_DATABASES
    assert isinstance(screen.focused_widget, DBView)

def test_screen_shows_sql_error(sakila_connection):
    screen = session.Session(sakila_connection)
    screen._state_machine.change_state('run_query', None, 'INVALID SQL')
    assert screen.view.showing_error_modal is True



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

class BaseWidgetsFactory:
    def __init__(self, state_machine):
        self._state_machine = state_machine

        # Contains signal name, handler and parameters for each created widget
        # Used by the destructor to disconnect any event handlers
        self._signal_handlers = []

        # Widgets cache
        self._widgets = {}
        self._factory_methods = self._get_factory_methods()

    def _get_factory_methods(self):
        raise NotImplementedError()

    def __del__(self):
        while True:
            try:
                obj, name, callback, kwargs = self._signal_handlers.pop()
                urwid.disconnect_signal(obj, name, callback, **kwargs)
            except IndexError:
                break

    def change_fsm_state(self, input, *args, **kwargs):
        self._state_machine.change_state(input, *args, **kwargs)

    def create(self, name, *args, **kwargs):
        cache_widget = True
        if 'cache' in kwargs:
            cache_widget = kwargs['cache']
            del kwargs['cache']

        if name not in self._widgets or cache_widget is False:
            factory_method = self._factory_methods[name]
            widget = factory_method(*args, **kwargs)

            if cache_widget is False:
                return widget
            self._widgets[name] = widget
        else:
            # Refresh the widget if it was cached
            widget = self._widgets[name]
            refresh_method = getattr(widget, 'refresh', None)
            if callable(refresh_method):
                refresh_method(*args, **kwargs)
        return widget

    def _connect_signal(self, obj, name, callback, **kwargs):
        key = urwid.connect_signal(obj, name, callback, **kwargs)
        self._signal_handlers.append((obj, name, callback, kwargs))




from mitzasql.state_machine import StateMachine
from mitzasql.logger import logger

# initial state: parse_char
#   if char is ' or '
#       goto state 'is quote'
#   else if char is [space] or end of string
#       goto state 'end'
#   else
#       goto state parse_char

class StopParsing(Exception):
    pass

class Parser:
    def __init__(self, string):
        self.string = string;
        self._tokens = []
        self._token_index = 0
        self._index = 0
        self._quote_is_open = False
        self._fsm = StateMachine('table_filter_parser', disable_logging=True)
        self._init_fsm()
        self._fsm.set_initial_state('parse')

    def parse(self):
        try:
            self._fsm.run(self._index)
        except StopParsing:
            pass
        return self.tokens

    @property
    def tokens(self):
        return list(filter(lambda token: len(token) > 0, self._tokens))

    def handle_char(self, index):
        try:
            char = self.string[index]
        except IndexError as e:
            char = None

        if char is None:
            return self._fsm.change_state('stop')

        if char == u' ':
            return self._fsm.change_state('found_space')

        if char == u"'":
            return self._fsm.change_state('found_quote', "'")

        if char == u'"':
            return self._fsm.change_state('found_quote', '"')

        if char == u'\\':
            return self._fsm.change_state('found_backslash')

        self._append_char(char)
        self._index += 1
        return self._fsm.change_state('next_char', self._index)

    def handle_space(self):
        if not self._quote_is_open:
            self._tokens.append(u'')
            self._token_index += 1
        else:
            self._append_char(u' ')
        self._index += 1
        return self._fsm.change_state('next_char', self._index)

    def handle_quote(self, quote_char):
        if self._quote_is_open != False:
            if self._quote_is_open == quote_char:
                self._quote_is_open = False
            else:
                self._append_char(quote_char)
            self._index += 1
            return self._fsm.change_state('next_char', self._index)

        self._quote_is_open = quote_char
        self._index += 1
        return self._fsm.change_state('next_char', self._index)

    def handle_backslash(self):
        try:
            next_char = self.string[self._index + 1]
        except IndexError:
            self._append_char(u'\\')
            self._fsm.change_state('stop')

        if next_char == self._quote_is_open:
            self._append_char(next_char)
            self._index += 2
        else:
            self._append_char(u'\\')
            self._index += 1
        return self._fsm.change_state('next_char', self._index)

    def handle_stop(self):
        raise StopParsing()

    def _init_fsm(self):
        self._fsm.add_state('parse', self.handle_char, {
            'found_space': 'on_space',
            'found_quote': 'on_quote',
            'found_backslash': 'on_backslash',
            'next_char': 'parse',
            'stop': 'on_stop'
            })
        self._fsm.add_state('on_space', self.handle_space, {
            'next_char': 'parse',
            })
        self._fsm.add_state('on_quote', self.handle_quote, {
            'next_char': 'parse',
            })
        self._fsm.add_state('on_backslash', self.handle_backslash, {
            'next_char': 'parse',
            })
        self._fsm.add_state('on_stop', self.handle_stop)

    def _append_char(self, char):
        try:
            token = self._tokens[self._token_index]
        except IndexError as e:
            self._tokens.append(u'')

        self._tokens[self._token_index] += char



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.widgets.info_widget import InfoWidget

class ProcedureWidget(InfoWidget):
    def __init__(self, model):
        self._model = model;
        head_section = self._create_head_section()
        body_section = self._create_body_section()

        contents = [head_section]
        contents.extend(body_section)
        super().__init__(contents)

    @property
    def name(self):
        return u'{0} {1}'.format(self._model['ROUTINE_TYPE'],
                self._model['SPECIFIC_NAME'])

    def _create_head_section(self):
        grid = []

        contents = []
        contents.append((18, urwid.AttrMap(urwid.Text(u'Name'),
            'editbox:label')))
        contents.append(urwid.AttrMap(urwid.Text(self._model['SPECIFIC_NAME']),
            'editbox'))
        grid.append(urwid.Columns(contents))

        contents = []
        contents.append((18, urwid.AttrMap(urwid.Text(u'Definer'),
            'editbox:label')))
        contents.append(urwid.AttrMap(urwid.Text(self._model['DEFINER']),
            'editbox'))
        grid.append(urwid.Columns(contents))

        contents = []
        contents.append((18, urwid.AttrMap(urwid.Text(u'Type'),
            'editbox:label')))
        contents.append(urwid.AttrMap(urwid.Text(self._model['ROUTINE_TYPE']),
            'editbox'))
        grid.append(urwid.Columns(contents))

        contents = []
        contents.append((18, urwid.AttrMap(urwid.Text(u'Data access'),
            'editbox:label')))
        contents.append(urwid.AttrMap(urwid.Text(self._model['SQL_DATA_ACCESS']),
            'editbox'))
        grid.append(urwid.Columns(contents))

        if (self._model['ROUTINE_TYPE'] == 'FUNCTION'):
            contents = []
            contents.append((18, urwid.AttrMap(urwid.Text(u'Returns'),
                'editbox:label')))
            contents.append(urwid.AttrMap(urwid.Text(self._model['DTD_IDENTIFIER']),
                'editbox'))
            grid.append(urwid.Columns(contents))

        contents = []
        contents.append((18, urwid.AttrMap(urwid.Text(u'SQL Security'),
            'editbox:label')))
        contents.append(urwid.AttrMap(urwid.Text(self._model['SECURITY_TYPE']),
            'editbox'))
        grid.append(urwid.Columns(contents))

        contents = []
        contents.append((18, urwid.AttrMap(urwid.Text(u'Is deterministic?'),
            'editbox:label')))
        contents.append(urwid.AttrMap(urwid.Text(self._model['IS_DETERMINISTIC']),
            'editbox'))
        grid.append(urwid.Columns(contents))

        grid = urwid.GridFlow(grid, cell_width=40, h_sep=1, v_sep=1,
                align='left')
        return grid

    def _create_body_section(self):
        body = []
        body.append(urwid.Divider('-'))
        body.append(urwid.AttrMap(urwid.Text('Params'), 'default'))

        if not isinstance(self._model['param_list'], str):
            param_list = self._model['param_list'].decode(encoding='utf8').split(',')
        else:
            param_list = self._model['param_list'].split(',')
        for param in param_list:
            body.append(urwid.AttrMap(urwid.Text(param.strip()), 'editbox'))
        body.append(urwid.Divider('-'))

        if len(self._model['ROUTINE_COMMENT']):
            comment = self._model['ROUTINE_COMMENT']
            if not isinstance(comment, str):
                comment = comment.decode(encoding='utf8')
            body.append(urwid.AttrMap(urwid.Text(u'Comment: ' + comment),
                'default'))
            body.append(urwid.Divider('-'))
        body.append(urwid.AttrMap(urwid.Text(self._model['ROUTINE_DEFINITION']),
            'default'))
        body.append(urwid.Text(''))
        return body



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from collections import deque
import urwid

from mitzasql.ui.widgets.base_db_view import BaseDBView
from mitzasql.ui.widgets.mysql_table import MysqlTable
from mitzasql.table_filter_parser import Parser
from mitzasql.logger import logger

from mitzasql.ui.widgets.cmd_proc import (BaseCmdProcessor, CommandError)

class CommandProcessor(BaseCmdProcessor):
    def __init__(self, table_view):
        self._table_view = table_view
        BaseCmdProcessor.__init__(self)

        self._colon_cmds.extend([
                ('r', 'resize', self._resize_table),
                ('s', 'sort', self._sort_table),
                ('e', 'eq', self._filter_table('eq')),
                ('ne', 'neq', self._filter_table('neq')),
                ('lt', 'lt', self._filter_table('lt')),
                ('lte', 'lte', self._filter_table('lte')),
                ('gt', 'gt', self._filter_table('gt')),
                ('gte', 'gte', self._filter_table('gte')),
                ('in', 'in', self._filter_table('in')),
                ('nin', 'nin', self._filter_table('nin')),
                ('bt', 'between', self._filter_table('between')),
                ('nbt', 'nbetween', self._filter_table('nbetween')),
                ('n', 'null', self._filter_table('null')),
                ('nn', 'nnull', self._filter_table('nnull')),
                ('empty', 'empty', self._filter_table('empty')),
                ('nempty', 'nempty', self._filter_table('nempty')),
                ('k', 'like', self._filter_table('like')),
                ('nk', 'nlike', self._filter_table('nlike')),
                ('cf', 'clearfilters', self._filter_table('clearfilters'))
                ])

        self._cmd_strings = {
                ':': self.colon_handler
                }

    def _resize_table(self, *args):
        if not len(args):
            return
        if len(args) == 1:
            column = args[0]
            value = 1
        else:
            column, *value = args
            value = ' '.join(value)
        try:
            self._table_view.resize_tbl_col(column, value)
        except CommandError as e:
            self._emit_error(str(e))

    def _filter_table(self, op):
        def callback(*args):
            try:
                self._table_view.filter(op, *args)
            except CommandError as e:
                self._emit_error(str(e))
        return callback

    def _sort_table(self, *args):
        if not len(args):
            return

        if len(args) == 1:
            column = args[0]
            direction = 'asc'
        else:
            column, *direction = args
            direction = ' '.join(direction)
        try:
            self._table_view.sort(column, direction)
        except CommandError as e:
            self._emit_error(str(e))

class TableView(BaseDBView):
    SIGNAL_ACTION_SELECT_ROW = 'select_row'
    SIGNAL_ACTION_INFO = 'info'
    SIGNAL_ACTION_CHANGE_TABLE = 'change_table'

    def __init__(self, model, connection):
        self._command_processor = CommandProcessor(self)
        self._table_widget_cls = MysqlTable
        actions = [
                ('F3', u'F3 Info', self.SIGNAL_ACTION_INFO)
                ]
        super().__init__(model, connection, actions)
        self._set_cmd_args_suggestions()
        self.SIGNALS.extend([self.SIGNAL_ACTION_SELECT_ROW,
            self.SIGNAL_ACTION_INFO, self.SIGNAL_ACTION_CHANGE_TABLE])
        self._connect_table_signals()

    def refresh(self, table, connection):
        if table == self._model.table_name:
            return

        self._model.table_name = table
        self._model.reload()
        self._set_cmd_args_suggestions()
        self._update_breadcrumbs()

    def refresh_model(self, emitter, action):
        self._model.reload(reset_limit=False, reset_order=False, reset_where=False)

    def _set_cmd_args_suggestions(self):
        cmd_args_suggestions = [c['name'] for c in self._model.columns]
        self._command_processor.cmd_args_suggestions['eq'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['neq'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['lt'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['lte'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['gt'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['gte'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['in'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['nin'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['between'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['nbetween'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['null'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['nnull'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['empty'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['nempty'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['like'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['nlike'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['resize'] = cmd_args_suggestions
        self._command_processor.cmd_args_suggestions['sort'] = self._command_processor.cmd_args_suggestions['resize']


    def __del__(self):
        self._disconnect_table_signals()
        super().__del__()

    def _connect_table_signals(self):
        urwid.connect_signal(self._table, self._table.SIGNAL_ROW_SELECTED,
                self.select_row)

    def _disconnect_table_signals(self):
        urwid.disconnect_signal(self._table, self._table.SIGNAL_ROW_SELECTED,
                self.select_row)

    def select_row(self, emitter, row):
        urwid.emit_signal(self, self.SIGNAL_ACTION_SELECT_ROW, self, row)

    @property
    def table(self):
        return self._model.table_name

    def resize_tbl_col(self, column, value=1):
        try:
            value = int(value)
            col_index = self._model.col_index(column.lower())
        except IndexError as e:
            raise CommandError(u'Column not found!')
        except ValueError as e:
            raise CommandError(u'Invalid column width!')
        self._table.resize_col(col_index, value)

    def sort(self, column, direction):
        if column not in [c['name'] for c in self._model.columns]:
            raise CommandError(u'Column not found!')

        if direction.lower() != 'asc' and direction.lower() != 'desc':
            raise CommandError(u'Invalid sort direction!')

        self._model.sort(column, direction)

    def filter(self, op, *args):
        if op == 'clearfilters':
            self._model.filter(None)
            return

        if not len(args):
            raise CommandError(u'Filter requires column!')

        arguments = deque(args)
        column = arguments.popleft()
        if column not in [c['name'] for c in self._model.columns]:
            raise CommandError(u'Column not found!')

        filters_without_values = ['null', 'nnull', 'empty', 'nempty']
        if not op in filters_without_values:
            filter_val = ' '.join(arguments)
            if not len(filter_val):
                raise CommandError("Filter requires value")

        if op == 'null':
            where = '`{0}` IS NULL'.format(column)
        elif op == 'nnull':
            where = '`{0}` IS NOT NULL'.format(column)
        elif op == 'empty':
            where = '`{0}` = \'\''.format(column)
        elif op == 'nempty':
            where = '`{0}` != \'\''.format(column)
        elif op == 'eq':
            where = '`{0}` = {1}'.format(column, filter_val)
        elif op == 'neq':
            where = '`{0}` != {1}'.format(column, filter_val)
        elif op == 'lt':
            where = '`{0}` < {1}'.format(column, filter_val)
        elif op == 'lte':
            where = '`{0}` <= {1}'.format(column, filter_val)
        elif op == 'gt':
            where = '`{0}` > {1}'.format(column, filter_val)
        elif op == 'gte':
            where = '`{0}` >= {1}'.format(column, filter_val)
        elif op == 'like':
            where = '`{0}` LIKE "{1}"'.format(column, filter_val)
        elif op == 'nlike':
            where = '`{0}` NOT LIKE "{1}"'.format(column, filter_val)
        elif op == 'in':
            values = ','.join(map(str.strip, filter_val.split(',')))
            if not len(values):
                raise CommandError("Filter requires value")
            where = '`{0}` IN ({1})'.format(column, values)
        elif op == 'nin':
            values = ','.join(map(str.strip, filter_val.split(',')))
            if not len(values):
                raise CommandError("Filter requires value")
            where = '`{0}` NOT IN ({1})'.format(column, values)
        elif op == 'between':
            p = Parser(filter_val)
            values = p.parse()
            if len(values) != 2:
                raise CommandError("Filter requires 2 values")
            values = list(map(lambda v: v.replace("'", "\\'"), values))
            where = "{0} BETWEEN '{1}' AND '{2}'".format(column, values[0], values[1])
        elif op == 'nbetween':
            p = Parser(filter_val)
            values = p.parse()
            if len(values) != 2:
                raise CommandError("Filter requires 2 values")
            values = list(map(lambda v: v.replace("'", "\\'"), values))
            where = "{0} NOT BETWEEN '{1}' AND '{2}'".format(column, values[0], values[1])

        self._model.filter(where)

    def keypress(self, size, key):
        if key == 'ctrl o':
            urwid.emit_signal(self, self.SIGNAL_ACTION_CHANGE_TABLE, self)
            return
        return super().keypress(size, key)



import pytest

from mitzasql.db.model import ProcedureModel
from .connection_fixture import sakila_connection

def test_model_fetches_data(sakila_connection):
    model = ProcedureModel(sakila_connection, 'sakila', 'rewards_report')

    assert model.last_error is None
    assert isinstance(model.data, dict)

    assert model['ROUTINE_SCHEMA'] == 'sakila'
    assert model['SPECIFIC_NAME'] == 'rewards_report'
    assert model['ROUTINE_TYPE'] == 'PROCEDURE'
    assert len(model['ROUTINE_DEFINITION']) > 0

def test_model_is_empty(sakila_connection):
    model = ProcedureModel(sakila_connection, 'sakila', 'unknown function')

    assert model.last_error is None
    assert len(model) == 0

def test_model_does_not_raise_exception_on_error(sakila_connection):
    model = ProcedureModel(sakila_connection, 'sakila\' FROM x',
            'rewards_report')

    assert model.last_error is not None



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
from mitzasql.sessions_registry import SessionsRegistry
from mitzasql.logger import LoggerMixin
from mitzasql.ui import ui
from mitzasql.db.connection import Connection
import mitzasql.constants as const

class App(LoggerMixin):
    """Main application class

    Initializes the UI and loads the appropriate screen based on the parameters
    passed from the main script.
    """
    def __init__(self, session_name=None, sessions_file=None):
        ui.init()
        self._session_name = session_name;
        self._sessions_registry = SessionsRegistry.get_instance(sessions_file)
        self._screen = None
        self.connection = None
        self.set_log_prefix('APP')
        self._set_screen()

    def start_ui(self, macro_file=None):
        """Start the UI loop. If a macro file was given run the macro

        Parameters:
        macro_file:string Path to macro file
        """
        ui.start_loop(macro_file)

    def _set_screen(self):
        if self._session_name is None:
            self.log_debug('Showing the "select session" screen')
            screen = ui.build_screen(const.SCREEN_SESSION_SELECT, display=True,
                    **{'sessions_registry': self._sessions_registry})
            ui.connect_signal(screen, screen.SIGNAL_CONNECT,
                    self.create_connection)
            ui.connect_signal(screen, screen.SIGNAL_TEST_CONNECTION,
                    self.test_connection)
            self._screen = screen
            return;

        self.log_debug("Showing the database view")
        connection_data = self._sessions_registry[self._session_name]

        con, error_message = Connection.factory(connection_data,
                session_name=self._session_name)

        if error_message is not None:
            print("Unable to connect: {0}".format(error_message),
                    file=sys.stderr)
            sys.exit(1)

        screen = ui.build_screen(const.SCREEN_SESSION, display=True,
            connection=con)
        self._screen = screen

    def create_connection(self, select_session_screen, session_name=None,
            connection_data=None):
        """Create database connection and load the appropriate screen

        If `session_name` is valid, get the connection details from the
        registry, or else use the details from `connection_data`

        Parameters:
        select_session_screen:ui.screens.session_select.SessionSelect The screen that triggered the connect signal
        session_name:String The session name
        connection_data:dict The connection details

        """
        select_session_screen.show_message(u'Connecting...')
        ui.refresh() # required because the next call will block the event loop

        if session_name is not None:
            connection_data = self._sessions_registry[session_name]

        con, error_message = Connection.factory(connection_data,
                session_name=session_name)
        if con is None:
            select_session_screen.set_connection_result(False,
                    str(error_message), session_name, connection_data)
            return

        self.connection = con
        screen = ui.build_screen(const.SCREEN_SESSION, display=True,
            connection=con)
        self._screen = screen

    def test_connection(self, select_session_screen, connection_data):
        """Test database connection

        Try to connect to the server using `connection_data`

        Paramters:
        select_session_screen:ui.screens.session_select.SessionSelect The screen that triggered the test signal
        connection_data:dict The connection details
        """

        select_session_screen.show_message(u'Connecting...')
        ui.refresh() # required because the next call will block the event loop
        result, error_message = Connection.test(connection_data)

        if result is False:
            select_session_screen.set_connection_test_result(False,
                    error=error_message)
            return
        select_session_screen.set_connection_test_result(True)

    @classmethod
    def run(cls, session_name=None, macro=None, sessions_file=None):
        """Run the application"""
        instance = cls(session_name, sessions_file=sessions_file)
        instance.start_ui(macro)
        # Cleanup
        if instance.connection:
            instance.connection.close()



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.screens.screen import Screen
from mitzasql.ui.widgets.saved_sessions_list import SavedSessionsList
from mitzasql.ui.widgets.edit_session_form import EditSessionForm
from mitzasql.ui.widgets.create_new_session_dialog import CreateNewSessionDialog

class SessionSelect(Screen):
    """The session select UI screen

    Implements the logic for creating, editing, deleting and selecting the
    session to connect to. When the user selects a session to connect to, the
    screen calls the main application instance with the session name or
    connection details.
    """

    # Signals handled by the main application instance
    SIGNAL_TEST_CONNECTION = 'test_connection'
    SIGNAL_CONNECT = 'connect'

    def __init__(self, sessions_registry):
        super().__init__([self.SIGNAL_CONNECT, self.SIGNAL_TEST_CONNECTION])
        self._sessions_registry = sessions_registry
        self._edit_form_widget = None
        self._sessions_list = None
        self.show_inital_widget()

    def __del__(self):
        self.focused_widget = None
        self._edit_form_widget = None
        self.view = None

    def show_inital_widget(self, *args):
        if self._sessions_registry.is_empty():
            return self._show_create_new_session_dialog()

        return self.show_sessions_list()

    def show_sessions_list(self, *args):
        sessions = self._sessions_registry.sessions
        if not self._sessions_list:
            list_ = SavedSessionsList(sessions)
            urwid.connect_signal(list_, SavedSessionsList.SIGNAL_QUIT, self.quit)
            urwid.connect_signal(list_, SavedSessionsList.SIGNAL_CREATE_NEW, self.create_new_session)
            urwid.connect_signal(list_, SavedSessionsList.SIGNAL_CONNECT, self.connect)
            urwid.connect_signal(list_, SavedSessionsList.SIGNAL_EDIT, self.edit_session)
            urwid.connect_signal(list_, SavedSessionsList.SIGNAL_DELETE, self.delete_session)
            self._sessions_list = list_
        else:
            self._sessions_list.refresh(sessions)

        self.focused_widget = self._sessions_list
        self.view.original_widget = urwid.Filler(self.focused_widget)

    def _show_create_new_session_dialog(self, *args, **kwargs):
        dialog = CreateNewSessionDialog()

        urwid.connect_signal(dialog, CreateNewSessionDialog.SIGNAL_YES, self.show_edit_session_form)
        urwid.connect_signal(dialog, CreateNewSessionDialog.SIGNAL_QUIT, self.quit)

        self.focused_widget = dialog
        self.view.original_widget = urwid.Filler(urwid.Padding(self.focused_widget, align='center',
            width=('relative', 50)))

    def show_edit_session_form(self, *args):
        if not self._edit_form_widget:
            form = EditSessionForm()
            urwid.connect_signal(form, form.SIGNAL_SAVE, self.save_connection)
            urwid.connect_signal(form, form.SIGNAL_TEST, self.test_connection)
            urwid.connect_signal(form, form.SIGNAL_CONNECT, self.connect)
            urwid.connect_signal(form, form.SIGNAL_CANCEL, self.show_inital_widget)
            self._edit_form_widget = form

        self._edit_form_widget.clear_status_message()
        self._edit_form_widget.refresh({})
        self._edit_form_widget.reset_focus()
        self.focused_widget = self._edit_form_widget
        self.view.original_widget = urwid.Filler(self.focused_widget)

    def create_new_session(self, *args):
        self.show_edit_session_form()
        self._edit_form_widget.refresh({}, True)

    def edit_session(self, *args):
        session_name = self._sessions_list.session
        form_data = self._sessions_registry[session_name]
        form_data['name'] = session_name
        self.show_edit_session_form()
        self._edit_form_widget.refresh(form_data, False)

    def save_connection(self, *args):
        form = self._edit_form_widget
        form_data = form.data
        validation_error = self._validate_connection_data(form_data)

        if validation_error is not None:
            form.set_status_message(validation_error, error=True)
            return

        if 'old_name' in form_data and form_data['old_name'] is not None:
            # Handle update session name
            old_name = form_data['old_name']
            del form_data['old_name']
            self._sessions_registry.edit(old_name, form_data)
        else:
            if 'old_name' in form_data and form_data['old_name'] is None:
                del form_data['old_name']
            self._sessions_registry.add(form_data)
        self._sessions_registry.save()
        self.show_sessions_list()

    def test_connection(self, *args):
        form = self._edit_form_widget
        form_data = form.data
        validation_error = self._validate_connection_data(form_data)

        if validation_error is not None:
            form.set_status_message(validation_error, error=True)
            return

        form.clear_status_message()
        urwid.emit_signal(self, self.SIGNAL_TEST_CONNECTION, self, form_data)

    def connect(self, *args):
        widget, *args = args

        if isinstance(widget, SavedSessionsList):
            session_name = widget.session
            urwid.emit_signal(self, self.SIGNAL_CONNECT, self, session_name)
            return

        if isinstance(widget, EditSessionForm):
            data = widget.data
            validation_error = self._validate_connection_data(data)
            if validation_error is not None:
                widget.set_status_message(validation_error, error=True)
                return
            urwid.emit_signal(self, self.SIGNAL_CONNECT, self, None, data)
            return

        raise RuntimeError('The current widget should not trigger a connect signal: {0}'.format(str(widget)))

    def delete_session(self, emitter, action):
        session_name = emitter.session
        del self._sessions_registry[session_name]
        self._sessions_registry.save()
        self.show_inital_widget()

    def show_message(self, message):
        widget = self.focused_widget
        widget.set_status_message(message)

    def set_connection_test_result(self, result, error=None):
        form = self._edit_form_widget
        if result is False:
            form.set_status_message(error, error=True)
            return

        form.set_status_message(u'Connection was successful')

    def set_connection_result(self, result, error=None, session_name=None, connection_data=None):
        if result is False:
            self.focused_widget.set_status_message(error, error=True)
        # nothing to do if result is True, our job is done

    def _validate_connection_data(self, data):
        if len(data['name']) == 0:
            return u'Session name is required'

        if len(data['host']) == 0:
            return u'Host is required'

        host = data['host']
        if not host.startswith('tcp://') and not host.startswith('unix://'):
            return u'Host is missing protocol (tcp://, unix://)'

        host_is_tcp = False

        if host.startswith('tcp://'):
            host_is_tcp = True
            host = host[6:]
        elif host.startswith('unix://'):
            host = host[7:]

        if len(host) == 0:
            return u'Host must not be empty'

        if host_is_tcp is True and ('port' not in data or len(data['port']) == 0):
            return u'Port is required'

        if 'username' not in data or len(data['username']) == 0:
            return u'Username is required'

        return None



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""urwid palette settings"""
default_theme = [
        ('error_message', 'white', 'dark red'),
        ('warning_message', 'black', 'brown'),
        ('info_message', 'white', 'dark blue', '', 'h186', 'g20'),

        ('action_bar', 'black', 'light gray', '', 'h250', 'g20'),
        ('action_bar:action', 'white', 'dark blue', '', 'g80', 'g30'),
        ('action_bar:action_key', 'light red,bold', 'dark blue', '', 'h122,bold', 'g30'),
        # ('action_bar:content', 'light gray', '', '', 'light gray', ''),
        ('action_bar:content', '', '', '', '', ''),

        ('linebox', '', '', '', 'h186', ''),

        ('session_header', 'light red', '', '', 'h186,bold', 'g20'),

        ('command_bar', 'white', 'dark blue', '', 'h117', 'g20'),

        ('list:item:unfocused', '', '', 'standout'),
        ('list:item:focused', 'black', 'light gray', 'standout'),

        ('editbox', 'white', 'dark blue', '', 'g20,bold', 'h144'),
        ('editbox:label', '', ''),

        ('theader', 'light cyan', ''),
        ('theader_sep', 'dark cyan', ''),

        ('trow_focused', 'black', 'light gray'),
        ('tcell_more', 'light red', ''),

        ('tcell_null_unfocused', 'dark red', ''),
        ('tcell_null_focused', 'dark red', 'light gray'),

        ('tfooter', 'light red', '', '', 'h186', 'g10')

        ]

themes = {
        'default': default_theme
        }

def get_palette(name = 'default'):
    return themes[name]




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.utils import orig_w
from .action_bar_pile_container import ActionBarPileContainer
from .emacs_edit import EmacsEdit
from .emacs_intedit import EmacsIntEdit

class EditSessionForm(ActionBarPileContainer):
    """Create & edit sessions"""

    # Signals handled by the select session UI screen
    SIGNAL_SAVE = 'save'
    SIGNAL_TEST = 'test'
    SIGNAL_CONNECT = 'connect'
    SIGNAL_CANCEL = 'cancel'

    def __init__(self, data={}, editing_new_session=True):

        actions = [
            ('F1', u'F1 Save', self.SIGNAL_SAVE),
            ('F2', u'F2 Test connection', self.SIGNAL_TEST),
            ('F3', u'F3 Connect', self.SIGNAL_CONNECT),
            ('Esc', u'Esc Cancel', self.SIGNAL_CANCEL)
            ]

        self.editing_new_session = editing_new_session
        self._old_session_name = None

        if len(data) == 0:
            self._data = {
                    'name': '',
                    'host': '',
                    'port': '',
                    'username': '',
                    'password': '',
                    'database': ''
                    }
        else:
            self._data = data
        self._input_elements = {}
        self._focus_position = 0
        self._name_field_container = None
        self._form = self._create_form()

        super().__init__(self._form, actions, line_box=True,
                line_box_title=u'Create Session')

    @property
    def title(self):
        return self._container.get_title()

    @title.setter
    def title(self, title):
        self._container.set_title(title)

    def _create_form(self):
        if not self.editing_new_session:
            name_is_read_only = True
            name_attr = 'editbox:label'
        else:
            name_is_read_only = False
            name_attr = 'editbox'

        name_label = urwid.Text(u'Session name:')
        name_edit = EmacsEdit(edit_text=self['name'],
                read_only=name_is_read_only)
        name_edit.value_name = 'name'
        self._input_elements['name'] = name_edit
        urwid.connect_signal(name_edit, 'postchange', self.on_input_change)
        name_edit = urwid.AttrMap(name_edit, name_attr)

        host_label = urwid.Text(u'Host:')
        host_edit = EmacsEdit(edit_text=self['host'])
        host_edit.value_name = 'host'
        self._input_elements['host'] = host_edit
        urwid.connect_signal(host_edit, 'postchange', self.on_input_change)
        host_edit = urwid.AttrMap(host_edit, 'editbox')

        port_label = urwid.Text(u'Port:')
        port_edit = EmacsIntEdit(default=self['port'])
        port_edit.value_name = 'port'
        self._input_elements['port'] = port_edit
        urwid.connect_signal(port_edit, 'postchange', self.on_input_change)
        port_edit = urwid.AttrMap(port_edit, 'editbox')

        username_label = urwid.Text(u'Username:')
        username_edit = EmacsEdit(edit_text=self['username'])
        username_edit.value_name = 'username'
        self._input_elements['username'] = username_edit
        urwid.connect_signal(username_edit, 'postchange', self.on_input_change)
        username_edit = urwid.AttrMap(username_edit, 'editbox')

        password_label = urwid.Text(u'Password:')
        password_edit = EmacsEdit(edit_text=self['password'], mask='*')
        password_edit.value_name = 'password'
        self._input_elements['password'] = password_edit
        urwid.connect_signal(password_edit, 'postchange', self.on_input_change)
        password_edit = urwid.AttrMap(password_edit, 'editbox')

        database_label = urwid.Text(u'Database:')
        database_edit = EmacsEdit(edit_text=self['database'])
        database_edit.value_name = 'database'
        self._input_elements['database'] = database_edit
        urwid.connect_signal(database_edit, 'postchange', self.on_input_change)
        database_edit = urwid.AttrMap(database_edit, 'editbox')

        contents = []

        if not name_is_read_only:
            row = urwid.Columns([(15, name_label), (32, name_edit),
                urwid.Text(u'Required')], dividechars=1)
        else:
            row = urwid.Columns([(15, name_label), (32, name_edit)], dividechars=1)
        self._name_field_container = row
        contents.append(row)

        contents.append(urwid.Divider())

        row = urwid.Columns([(15, host_label), (32, host_edit),
            urwid.Text(u'Must start with tcp:// or unix://')], dividechars=1)
        contents.append(row)

        if name_is_read_only:
            self._focus_position = 2

        contents.append(urwid.Divider())

        row = urwid.Columns([(15, port_label), (6, port_edit),
            urwid.Text(u'Required only for tcp:// connections')], dividechars=1)
        contents.append(row)

        contents.append(urwid.Divider())

        row = urwid.Columns([(15, username_label), (32, username_edit),
            urwid.Text(u'Required')], dividechars=1)
        contents.append(row)

        contents.append(urwid.Divider())

        row = urwid.Columns([(15, password_label), (32, password_edit)],
                dividechars=1)
        contents.append(row)

        contents.append(urwid.Divider())

        row = urwid.Columns([(15, database_label), (32, database_edit)],
                dividechars=1)
        contents.append(row)

        contents.append(urwid.Divider())

        pile = urwid.Pile(contents)
        return pile

    def __getitem__(self, name):
        if name in self._data:
            return self._data[name]
        return ''

    def refresh(self, data, editing_new_session=None):
        self._data = data
        self.editing_new_session = editing_new_session
        for name, widget in self._input_elements.items():
            if name not in data:
                value = ''
            else:
                value = data[name]
            widget.edit_text = value
        self._old_session_name = None

        if editing_new_session == None:
            return

        if not editing_new_session:
            self._toggle_name_read_only(True)
            return

        self._toggle_name_read_only(False)

    def _toggle_name_read_only(self, state):
        name_field = self._name_field_container.contents[1][0]

        if orig_w(name_field).read_only == state:
            return

        if state:
            orig_w(name_field).read_only = True
            name_field.set_attr_map({None: 'editbox:label'})
            del self._name_field_container.contents[2]
            self._focus_position = 2
        else:
            orig_w(name_field).read_only = False
            name_field.set_attr_map({None: 'editbox'})

            if len(self._name_field_container.contents) == 2:
                options = self._name_field_container.options('pack')
                self._name_field_container.contents.append((urwid.Text(u'Required'), options))
                self._focus_position = 0
        self._name_field_container.focus_position = 1
        self.reset_focus()

    @property
    def data(self):
        data = self._data.copy()

        if self.editing_new_session is False:
            data['old_name'] = self._old_session_name

        return data

    def on_input_change(self, emitter, old_value):
        if emitter.value_name == 'name' and self.editing_new_session is False:
            if self._old_session_name is None and 'name' in self._data:
                self._old_session_name = self._data['name']
        self._data[emitter.value_name] = emitter.edit_text

    def keypress(self, size, key):
        if key == 'tab':
            key = 'down'
        elif key == 'shift tab':
            key = 'up'
        return super().keypress(size, key)

    def reset_focus(self):
        self._form.focus_position = self._focus_position




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

exec(open('mitzasql/version.py').read())

setuptools.setup(
        name='mitzasql',
        version=__version__,
        author='Vlad Balmos',
        author_email='vladbalmos@yahoo.com',
        long_description_content_type='text/markdown',
        description='Text user interface MySQL client',
        long_description=long_description,
        license="MIT",
        url='https://github.com/vladbalmos/mitzasql',
        project_urls={
            'Bug Tracker': 'https://github.com/vladbalmos/mitzasql/issues',
            'Website': 'https://vladbalmos.github.io/mitzasql'
        },
        packages=setuptools.find_packages(),
        keywords='tui cli ncurses console mysql client',
        package_data={
            'mitzasql': ['ui/widgets/help.txt']
        },
        classifiers=[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX :: Linux',
        ],
        install_requires=[
            'urwid ==2.0.1',
            'mysql-connector-python ==8.0.16',
            'appdirs ==1.4.3'
        ],
        scripts=['bin/mitzasql']
)



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from mitzasql.logger import LoggerMixin

class StateMachine(LoggerMixin):
    """Finite state machine implementation"""
    def __init__(self, name, disable_logging=False):
        self._name = name
        self._handlers = {}
        self._transition_table = {}
        self._current_state = None
        self._disable_logging = disable_logging
        self.set_log_prefix('FSM')

    def add_state(self, name, handler, transition_table=None):
        self._handlers[name] = handler
        if transition_table is not None:
            self._transition_table[name] = transition_table

    def get_current_state(self):
        return self._current_state

    def set_initial_state(self, name):
        self._log('Setting initial state to "%s"', name)
        self._current_state = name

    def run(self, *args, **kwargs):
        self._log('Transitioning to state "%s" with args: %s, kwargs: %s',
                self._current_state, args, kwargs)
        try:
            handler = self._handlers[self._current_state]
        except KeyError as e:
            self.log_critical('No handler registered for state: %s',
                    self._current_state)
            raise e

        handler(*args, **kwargs)
        self._log('Current state is "%s"', self._current_state)

    def change_state(self, input, *args, **kwargs):
        self._log('Received input: %s', input)
        try:
            transition_table = self._transition_table[self._current_state]
            next_state = transition_table[input]
        except KeyError as e:
            self.log_critical('Next state is not defined for input: %s', input)
            raise e

        self._current_state = next_state
        self.run(*args, **kwargs)

    def _log(self, message, *args):
        if self._disable_logging:
            return
        message = '"{0}": {1}'.format(self._name, message)
        self.log_debug(message, *args)



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from .action_bar import ActionBar

class ActionBarPileContainer(urwid.WidgetPlaceholder):
    """Container widget which wraps another widget in a Pile, togheter with a
    status and an action bar

    If a key associated with an action is pressed, the widget emits the corresponding action
    signal.
    """
    SIGNAL_ACTION = 'action'
    SIGNALS = [SIGNAL_ACTION]

    def __init__(self, body, actions, height=None, line_box=False, line_box_title=''):
        self._actions = []
        self._actions_to_signals = {}
        self._body = body
        self._status_bar = urwid.Text('')
        self._prepare_actions_and_signals(actions)
        self._action_bar = urwid.AttrMap(ActionBar(self._actions), 'action_bar')
        self._container = self._make_container(height=height)

        if line_box is True:
            self._container = urwid.LineBox(self._container,
                    title=line_box_title)

        super().__init__(urwid.AttrMap(self._container, 'linebox'))
        self._register_signals()

    def _make_container(self, height=None):
        if height is not None:
            body = (height, urwid.Filler(self._body))
        else:
            body = self._body

        container = urwid.Pile([self._body, ('weight', 1, self._status_bar), ('weight', 1, self._action_bar)])
        container = urwid.AttrMap(container, 'action_bar:content')
        return container

    def set_status_message(self, message, error=False, warning=False):
        if error is False and warning is False:
            attribute = 'info_message'
        elif error is True:
            attribute = 'error_message'
        elif warning is True:
            attribute = 'warning_message'

        self._status_bar.set_text((attribute, message))

    def clear_status_message(self):
        self._status_bar.set_text('')

    def _register_signals(self):
        signals = self.SIGNALS
        if len(self._actions_to_signals):
            signals.extend(self._actions_to_signals.values())
        urwid.register_signal(self.__class__, signals)

    def _prepare_actions_and_signals(self, actions):
        _actions = []
        actions_to_signals = {}
        for action in actions:
            if len(action) == 3:
                key, action, signal = action
            else:
                key, action = action
                signal = None

            _actions.append((key, action))

            if signal is not None:
                actions_to_signals[key] = signal

        self._actions = _actions
        self._actions_to_signals = actions_to_signals

    def selectable(self):
        """This is required in order to catch user input"""
        return True

    def keypress(self, size, key):
        # Pass the keypress to the body widget
        if self._body.selectable() is True and self._body.keypress(size, key) is None:
            return
        # Check if any of the unhandled keys are registered with any actions
        # and trigger any corresponding signals
        for action_key, action_name in self._actions:
            if action_key.lower() == key.lower():
                urwid.emit_signal(self, self.SIGNAL_ACTION, self, action_name)

                if action_key in self._actions_to_signals:
                    signal = self._actions_to_signals[action_key]
                    urwid.emit_signal(self, signal, self, action_name)
                return
        return key




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import configparser
import os

import mitzasql.constants as const

class SessionsRegistry:
    """
    Registry for MySQL connections settings.

    It reads the `sessions_file` for the previously saved sessions.
    Newly created sessions are written to the same file.
    """
    instance = None

    def __init__(self, sessions_file=None):
        self._sessions = configparser.ConfigParser()
        if sessions_file is None:
            self._sessions_file = const.DEFAULT_SESSIONS_FILE
        else:
            self._sessions_file = sessions_file

        if not os.path.isdir(os.path.dirname(self._sessions_file)):
            os.makedirs(os.path.dirname(self._sessions_file), mode=0o700,
                    exist_ok=True)
        self._sessions.read(self._sessions_file)

    @property
    def sessions(self):
        return self._sessions.sections()

    def has_session(self, name):
        return self._sessions.has_section(name)

    def create_session(self, name, settings):
        """Add a new session to the registry

        Parameters:
        name:String The name of the session
        settings:dict The connection details:
                        settings = {
                        host: String,
                        port: String,
                        username: String,
                        password: String,
                        database: String
                        }
        """
        string_settings = {key: str(value) for key, value in settings.items()}
        self._sessions[name] = string_settings

    def __getitem__(self, name):
        return dict(self._sessions.items(name))

    def is_empty(self):
        return len(self._sessions.sections()) == 0

    def save(self):
        with open(self._sessions_file, 'w') as sessions_file:
            self._sessions.write(sessions_file)

    def add(self, connection_data):
        name = connection_data['name'];
        del connection_data['name']
        self._sessions[name] = connection_data

    def edit(self, session_name, connection_data):
        new_name = connection_data['name']
        del connection_data['name']

        if session_name != new_name:
            del self._sessions[session_name]
            connection_data['name'] = new_name
            return self.add(connection_data)

        self._sessions[session_name] = connection_data

    def __delitem__(self, name):
        del self._sessions[name]

    @classmethod
    def get_instance(cls, sessions_file=None):
        if cls.instance is None:
            cls.instance = SessionsRegistry(sessions_file=sessions_file)
        return cls.instance




import os
import pytest

import urwid
import mysql.connector.errors as errors
from mitzasql.db.model import DBTablesModel
from .connection_fixture import connection

database = 'sakila'

def test_model_fetches_data(connection):
    model = DBTablesModel(connection, database)
    assert len(model) == 35
    assert 'Name' in [c['name'] for c in model.columns]

    table_names = []
    for row in model:
        table_names.append(row[0])

    assert 'actor' in table_names

def test_model_handles_error_if_database_is_invalid(connection):
    model = DBTablesModel(connection, database)

    assert model.last_error is None
    model.database = 'unknown database'
    model.reload()
    assert isinstance(model.last_error, errors.ProgrammingError)

def test_model_emits_signals_during_reloading(connection):
    model = DBTablesModel(connection, database)
    pre_load_signal_handled = False
    load_signal_handled = False

    def pre_load_handler(emitter):
        nonlocal pre_load_signal_handled
        pre_load_signal_handled = True

    def load_handler(emitter):
        nonlocal load_signal_handled
        load_signal_handled = True

    urwid.connect_signal(model, model.SIGNAL_PRE_LOAD, pre_load_handler)
    urwid.connect_signal(model, model.SIGNAL_LOAD, load_handler)

    model.reload()

    assert pre_load_signal_handled is True
    assert load_signal_handled is True




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.screens.session_select.session_select import SessionSelect
from mitzasql.ui.screens.session.session import Session
import mitzasql.constants as const

class Builder:
    """Screens builder

    Build UI screens and attaches their view to the top most widget.
    UI screens act like controllers for their child widgets and can interact
    directly with the main application instance.
    """

    def __init__(self, top_most_widget):
        self._top_most_widget = top_most_widget

        self._build_callbacks = {
                const.SCREEN_SESSION_SELECT: self._build_session_select,
                const.SCREEN_SESSION: self._build_session
                }

    def _build_session_select(self, sessions_registry):
        screen = SessionSelect(sessions_registry)
        return screen

    def _build_session(self, connection):
        screen = Session(connection)
        return screen

    def build_screen(self, name, display=False, **kwargs):
        build_callback = self._build_callbacks[name]
        screen = build_callback(**kwargs)
        if display is True:
            self._top_most_widget.original_widget = screen.view
        return screen




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui import main_loop as shared_main_loop
from mitzasql.ui.widgets.error_dialog import ErrorDialog
from mitzasql.ui.widgets.info_dialog import InfoDialog
from mitzasql.ui.widgets.quit_dialog import QuitDialog

class SessionPopupLauncher(urwid.PopUpLauncher):
    DEFAULT_WR = 60
    DEFAULT_HR = 30
    def __init__(self, widget):
        super().__init__(widget)
        self._max_width = None
        self._max_height = None
        self._width_ratio = self.DEFAULT_WR
        self._height_ratio = self.DEFAULT_HR
        # Used in testing
        self.showing_error_modal = False

        self._popup_factory_method = None

    def create_pop_up(self):
        return self._popup_factory_method()

    def get_pop_up_parameters(self):
        width = int(self._width_ratio * self._max_width / 100)
        height = int(self._height_ratio * self._max_height / 100)

        left = (self._max_width - width) / 2
        top = (self._max_height - height) / 2

        return {'left': left, 'top': top, 'overlay_width': width, 'overlay_height':
                height}

    def close_pop_up(self, *args, **kwargs):
        self.showing_error_modal = False
        self._reset_popup_size_ratio()
        return super().close_pop_up()

    def quit(self, *args, **kwargs):
        raise urwid.ExitMainLoop()

    def show_fatal_error(self, error):
        self.showing_error_modal = True
        def factory_method():
            dialog = ErrorDialog(error, title=u'Unhandled Exception',
                prefix='An unhandled error occured. Exception details:')
            urwid.connect_signal(dialog, dialog.SIGNAL_OK, self.quit)
            return urwid.Filler(dialog)
        self._popup_factory_method = factory_method
        return self.open_pop_up()

    def show_error(self, error):
        self.showing_error_modal = True
        def factory_method():
            dialog = ErrorDialog(error)
            urwid.connect_signal(dialog, dialog.SIGNAL_OK, self.close_pop_up)
            return urwid.Filler(dialog)
        self._popup_factory_method = factory_method
        return self.open_pop_up()

    def show_info(self, message):
        def factory_method():
            dialog = InfoDialog(message)
            urwid.connect_signal(dialog, dialog.SIGNAL_OK, self.close_pop_up)
            return urwid.Filler(dialog)
        self._popup_factory_method = factory_method
        return self.open_pop_up()

    def close_pop_up_then(self, callback):
        '''Close popup then execute callback'''
        def fn(*args, **kwargs):
            self.close_pop_up()
            callback()
        return fn

    def show_quit_dialog(self, on_no=None):
        def factory_method():
            dialog = QuitDialog()
            urwid.connect_signal(dialog, dialog.SIGNAL_OK, self.quit)

            if on_no is not None:
                urwid.connect_signal(dialog, dialog.SIGNAL_CANCEL,
                        self.close_pop_up_then(on_no))
            else:
                urwid.connect_signal(dialog, dialog.SIGNAL_CANCEL,
                        self.close_pop_up)
            return urwid.Filler(dialog)
        self._popup_factory_method = factory_method
        return self.open_pop_up()

    def show_loading_dialog(self):
        self._width_ratio = 40
        self._height_ratio = 30
        def factory_method():
            dialog = urwid.Text(u'\nLoading...', align='center')
            dialog = urwid.Filler(dialog)
            dialog = urwid.AttrMap(urwid.LineBox(dialog), 'linebox')
            return dialog
        self._popup_factory_method = factory_method
        result = self.open_pop_up()
        shared_main_loop.refresh()
        return result

    def show_big_popup(self, widget):
        self._width_ratio = 90
        self._height_ratio = 80
        def factory_method():
            dialog = urwid.AttrMap(urwid.LineBox(widget, title=widget.name), 'linebox')
            urwid.connect_signal(widget, widget.SIGNAL_ESCAPE, self.close_pop_up)
            urwid.connect_signal(widget, widget.SIGNAL_QUIT,
                    self.close_pop_up_then(self.show_quit_dialog))
            return dialog
        self._popup_factory_method = factory_method
        result = self.open_pop_up()
        shared_main_loop.refresh()
        return result

    def show_table_changer(self, widget):
        self._width_ratio = 25
        self._height_ratio = 70
        def factory_method():
            dialog = urwid.AttrMap(urwid.LineBox(widget, title=u'Change table'), 'linebox')
            urwid.connect_signal(widget, widget.SIGNAL_ESCAPE, self.close_pop_up)
            return dialog
        self._popup_factory_method = factory_method
        result = self.open_pop_up()
        shared_main_loop.refresh()
        return result

    def _reset_popup_size_ratio(self):
        self._width_ratio = self.DEFAULT_WR
        self._height_ratio = self.DEFAULT_HR

    def render(self, size, focus=False):
        self._max_width, self._max_height = size
        return super().render(size, focus)
